from qsynthesis.plugin.dependencies import ida_kernwin, TRITON_ENABLED, Trace


class SynthetizerViewHook(ida_kernwin.action_handler_t):
    """
    Standard hook to all Qtrace-IDA plugins. When used with
    Qtace-IDA, Qsynthesis is launched through this action_handler_t
    which will be registered in the qtrace-ida menu.
    """

    view_id = "QSynthesis"

    def __init__(self, qtrace: 'QtraceIDA'):
        """
        Constructor that receive a QtraceIDA instance (which is the main object of Qtrace-IDA
        holding the trace etc)
        """
        ida_kernwin.action_handler_t.__init__(self)
        from qtraceida.icons.raw_icons import ICON_DEBUG_ID, IDAIcon
        self.view = None
        self.name = "QSynthesis"  # SynthesizerView.NAME
        self.icon = IDAIcon.AST_TREE
        self.tooltip = "Synthesizing arithmetic expressions along the trace"
        self.qtrace = qtrace

    @property
    def is_enabled_by_trace(self) -> bool:
        """
        Boolean indicating if this qtraceida plugin should be enabled
        when a trace is opened.
        """
        return True

    def on_trace_opened(self, t: Trace) -> None:
        """
        Callback called when a trace is opened in Qtrace-IDA.
        Here we just forward the call to the main view.
        """
        if self.view:  # If the view has been instanciated
            self.view.on_trace_opened(t)

    def on_trace_closed(self) -> None:
        pass # TODO: Implementing proper deactivation of the view

    def activate(self, _) -> bool:
        """
        Main callback called when the menu entry is clicked or the Short-keys
        typed. This `action_handler_t` has to be implemented. If called the
        action open the QSynthesis view if not already done.
        """
        # pylint: disable=unused-argument
        from qsynthesis.plugin.view import SynthesizerView
        if self.view is None or self.view.closed:
            self.view = SynthesizerView(self.qtrace)
            self.view.Show()  # will call OnCreate and init and on_trace_opened by recursivity
        return True

    def update(self, _) -> int:
        """
        Callback called whenever the the hook is updated.
        If dependency are not satisfied always disabled otherwise ok.
        """
        if TRITON_ENABLED:
            return ida_kernwin.AST_DISABLE if self.is_enabled_by_trace else ida_kernwin.AST_ENABLE_ALWAYS
        else:
            self.tooltip = "Require Triton to run"
            return ida_kernwin.AST_DISABLE_ALWAYS



from enum import IntEnum
from triton import AST_NODE
from collections import namedtuple
import operator


class BoolOp(IntEnum):
    """
    Enum of SMT boolean operators using Triton AST_NODE
    enum value
    """
    # Bool x Bool -> Bool
    AND = AST_NODE.LAND
    NOT = AST_NODE.LNOT
    LOR = AST_NODE.LOR
    IFF = AST_NODE.IFF
    EQUAL = AST_NODE.EQUAL
    DISTINCT = AST_NODE.DISTINCT


class BvOp(IntEnum):
    """
    Enum of SMT Bitvector operators as declared by Triton AST_NODE
    """
    # Basic ops
    NOT = AST_NODE.BVNOT
    AND = AST_NODE.BVAND
    OR = AST_NODE.BVOR
    XOR = AST_NODE.BVXOR
    NEG = AST_NODE.BVNEG
    ADD = AST_NODE.BVADD
    MUL = AST_NODE.BVMUL
    SUB = AST_NODE.BVSUB
    SHL = AST_NODE.BVSHL
    LSHR = AST_NODE.BVLSHR
    ROL = AST_NODE.BVROL  # Int x Bv -> Bv
    ROR = AST_NODE.BVROR  # Int x Bv -> Bv
    UDIV = AST_NODE.BVUDIV
    UREM = AST_NODE.BVUREM
    ASHR = AST_NODE.BVASHR
    SDIV = AST_NODE.BVSDIV
    SREM = AST_NODE.BVSREM
    SMOD = AST_NODE.BVSMOD
    # Extended
    XNOR = AST_NODE.BVXNOR
    NOR = AST_NODE.BVNOR
    NAND = AST_NODE.BVNAND
    # Change bit size
    ZEXT = AST_NODE.ZX  # Int x Bv -> Bv
    SEXT = AST_NODE.SX  # Int x Bv -> Bv
    CONCAT = AST_NODE.CONCAT
    EXTRACT = AST_NODE.EXTRACT  # Int x Int x Bv -> Bv
    # Other: Bool x BV x BV -> Bv
    ITE = AST_NODE.ITE
    # Boolean ops: BV x BV -> Bool
    UGE = AST_NODE.BVUGE
    UGT = AST_NODE.BVUGT
    ULE = AST_NODE.BVULE
    ULT = AST_NODE.BVULT
    SLE = AST_NODE.BVSLE
    SLT = AST_NODE.BVSLT
    SGE = AST_NODE.BVSGE
    SGT = AST_NODE.BVSGT


Operator = namedtuple("Operator", "id symbol eval_trit arity commutative id_eq id_zero is_prefix can_overflow bool_ret")


OPERATORS = {               # ID               strop    Trit op            arit comm   id_eq  id_zero is_pfx  can_ov bool_ret
    # BoolOp.EQUAL:    Operator(BoolOp.EQUAL,    "==",    operator.eq,     2,   True,  True,  False,  False,  False, True),
    # BoolOp.DISTINCT: Operator(BoolOp.DISTINCT, "!=",    operator.ne,     2,   True,  False, True,   False,  False, True),
    # BoolOp.IFF:      Operator(BoolOp.IFF,      "iff",   "iff",           2,   False, False, False,  False,  False, True),
    # BoolOp.LOR:      Operator(BoolOp.LOR,      "lor",   "lor",           2,   True,  True,  False,  True,   False, True),
    # BoolOp.AND:      Operator(BoolOp.AND,      "land",  "land",          2,   True,  True,  False,  True,   False, True),
    # BoolOp.NOT:      Operator(BoolOp.NOT,      "lnot",  "lnot",          1,   False, False, False,  True,   False, True),
    BvOp.NOT:        Operator(BvOp.NOT,        "~",     operator.invert,   1,   False, False, False,  True,   False, False),
    BvOp.AND:        Operator(BvOp.AND,        "&",     operator.and_,     2,   True,  True,  False,  False,  False, False),
    BvOp.OR:         Operator(BvOp.OR,         '|',     operator.or_,      2,   True,  True,  False,  False,  False, False),
    BvOp.XOR:        Operator(BvOp.XOR,        '^',     operator.xor,      2,   True,  False, True,   False,  False, False),
    BvOp.NEG:        Operator(BvOp.NEG,        '-',     operator.neg,      1,   False, False, False,  True,   False, False),
    BvOp.ADD:        Operator(BvOp.ADD,        '+',     operator.add,      2,   True,  False, False,  False,  True,  False),
    BvOp.MUL:        Operator(BvOp.MUL,        '*',     operator.mul,      2,   True,  False, False,  False,  True,  False),
    BvOp.SUB:        Operator(BvOp.SUB,        '-',     operator.sub,      2,   False, False, True,   False,  False, False),
    BvOp.SHL:        Operator(BvOp.SHL,        "<<",    operator.lshift,   2,   False, False, False,  False,  True,  False),
    BvOp.LSHR:       Operator(BvOp.LSHR,       ">>",    operator.rshift,   2,   False, False, False,  False,   False, False),

    BvOp.ROL:        Operator(BvOp.ROL,        "bvrol", "bvrol",           2,   False, False, False,  True,   False, False),
    BvOp.ROR:        Operator(BvOp.ROR,        "bvror", "bvror",           2,   False, False, False,  True,   False, False),
    # BvOp.UDIV:       Operator(BvOp.UDIV,       "/",     operator.truediv,2,   False, False, False,  False,  False, False), # TODO: Fix /0
    # BvOp.UREM:       Operator(BvOp.UREM,       "%",     operator.mod,    2,   False, False, False,  False,  False, False), # TODO: Fix /0
    BvOp.ASHR:       Operator(BvOp.ASHR,       "bvashr", "bvashr",         2,   False, False, False,  True,   False, False),
    # BvOp.SDIV:       Operator(BvOp.SDIV,       "bvsdiv","bvsdiv",        2,   False, False, False,  False,  False, False),
    # BvOp.SREM:       Operator(BvOp.SREM,       "bvsrem","bvsrem",        2,   False, False, True,   True,   False, False),
    # BvOp.SMOD:       Operator(BvOp.SMOD,       "bvsmod","bvsmod",        2,   False, False, True,   False,  False, False),
    # BvOp.XNOR:       Operator(BvOp.XNOR,       "bvxnor","bvxnor",        2,   True,  False, False,  True,   False, False),
    # BvOp.NOR:        Operator(BvOp.NOR,        "bvnor", "bvnor",         2,   True,  False, False,  True,   False, False),
    # BvOp.NAND:       Operator(BvOp.NAND,       "bvnand","bvnand",        2,   True,  False, False,  True,   False, False),
    # BvOp.ZEXT:       Operator(BvOp.ZEXT,       "zx",    "zx",            2,   False, True,  False,  True,   False, False),
    # BvOp.SEXT:       Operator(BvOp.SEXT,       "sx",    "sx",            2,   False, False, False,  True,   False, False), # FIXME: operator
    # BvOp.CONCAT:     Operator(BvOp.CONCAT,     "concat","concat",        2,   False, False, False,  True,   False, False), # FIXME: operator
    # BvOp.EXTRACT:    Operator(BvOp.EXTRACT,    "extract","extract",      2,   False, False, False,  True,   False, False), # FIXME: operator
    # BvOp.ITE:        Operator(BvOp.ITE,        "If",     "If",           3,   False, False, False,  True,   False, False),
    # BvOp.UGE:        Operator(BvOp.UGE,        ">=",     operator.ge,    2,   False, True,  False,  False,  False, True),
    # BvOp.UGT:        Operator(BvOp.UGT,        ">",      operator.gt,    2,   False, False, True,   False,  False, True),
    # BvOp.ULE:        Operator(BvOp.ULE,        "<=",     operator.le,    2,   False, True,  False,  False,  False, True),
    # BvOp.ULT:        Operator(BvOp.ULT,        "<",      operator.lt,    2,   False, False, True,   False,  False, True),
    # BvOp.SLE:        Operator(BvOp.SLE,        "bvsle",  "bvsle",        2,   False, True,  False,  True,   False, True),
    # BvOp.SLT:        Operator(BvOp.SLT,        "bvslt",  "bvslt",        2,   False, False, True,   True,   False, True),
    # BvOp.SGE:        Operator(BvOp.SGE,        "bvsge",  "bvsge",        2,   False, True,  False,  True,   False, True),
    # BvOp.SGT:        Operator(BvOp.SGT,        "bvsgt",  "bvsgt",        2,   False, False, True,   True,   False, True)
}



# built-in libs
from pathlib import Path
import binascii

# third-party libs
import requests

# qsynthesis deps
from qsynthesis.grammar import TritonGrammar
from qsynthesis.tables.base import InputOutputOracle, HashType, Hash
from qsynthesis.types import Input, Optional, List, Dict, Union, Tuple, Iterator


class InputOutputOracleREST(InputOutputOracle):
    """
    REST-based lookup table. The table given in parameter is meant to be an IP
    address serving the API (that can be served with qsynthesis-table-manager runserver)
    """

    def __init__(self, grammar: TritonGrammar, inputs: List[Input], hash_mode: HashType = HashType.RAW, f_name: str = ""):
        """
        Constructor making a lookuptable from a grammar a set of inputs and an hash type.

        :param grammar: triton grammar
        :param inputs: List of inputs
        :param hash_mode: type of hash to be used as keys in tables
        :param f_name: file name of the table (when being loaded)
        """
        super(InputOutputOracleREST, self).__init__(grammar, inputs, hash_mode, f_name)
        self.session = requests.Session()
        self._size = 0

    @staticmethod
    def create(filename: Union[str, Path], grammar: TritonGrammar, inputs: List[Input], hash_mode: HashType = HashType.RAW, constants: List[int] = []) -> 'InputOutputOracleREST':
        """
        Such tables cannot be created as they are read-only databases.
        """
        raise RuntimeError("REST Lookup tables cannot be created only loaded")

    @staticmethod
    def load(file: Union[Path, str]) -> 'InputOutputOracleREST':
        """
        Load a given table. The function perform a request to retrieve alll
        the basic informations about the table (size, grammar, inputs etc..)
        and to create a completely transparent table.
        """
        res = requests.get(file)
        if res.status_code == 200:
            data = res.json()
            g = TritonGrammar.from_dict(data['grammar'])
            lkp = InputOutputOracleREST(g, data['inputs'], HashType[data['hash_mode']], file)
            lkp._size = data["size"]
            lkp.session.headers['Host'] = file
            lkp._name = file
            return lkp
        else:
            raise ConnectionAbortedError(f"Cannot reach remote server (code:{res.status_code})")

    def add_entry(self, hash: Hash, value: str):
        """
        Function not implemented at the moment. No new expressions can be
        submitted at the moment.
        """
        raise NotImplementedError("REST Lookup Table are read-only at the moment")

    def add_entries(self, entries: List[Tuple[Hash, str]], calc_hash: bool = False, chunk_size: int = 10000) -> None:
        """
        Function not implemented at the moment. REST tables are read-only.
        """
        raise NotImplementedError("REST Lookup tables are read-only at the moment")

    def __iter__(self) -> Iterator[Tuple[Hash, str]]:
        """
        Iterating over all elements of the table is not implemented either.
        """
        raise NotImplementedError("Entries iteration is not implemented")

    @property
    def size(self) -> int:
        """Size of the table (number of entries)"""
        return self._size

    def _get_item(self, h: Hash) -> Optional[str]:
        """
        From a given hash return the associated expression string if found in
        the lookup table through a GET request.

        :param h: hash of the item to get
        :returns: raw expression string if found
        """
        hex_hash = binascii.hexlify(h).decode()
        res = self.session.get(str(self.name) + "/entry/" + hex_hash)
        if res.status_code == 200:
            data = res.json()
            return data['expression'] if data else None
        else:
            raise ConnectionError("REST query did not suceeded correctly")



# Standard modules
from __future__ import annotations
from enum import IntEnum
import random
from functools import reduce
import logging

# Third-party modules
from triton import TritonContext, AST_NODE, SYMBOLIC, ARCH

# QSynthesis imports
from qsynthesis.types import AstNode, List, Tuple, Generator, Dict, Union, Optional, AstType, SymVarMap, \
                             SymbolicVariable, Char, IOVector, Input, Output


logger = logging.getLogger('qsynthesis')


class ReassemblyError(Exception):
    """
    Wrapping exception for all exceptions that might be raised during the
    reassembly process.
    """
    pass


class SymVarType(IntEnum):
    """
    Enum representing the different types of SymbolicVariables of Triton
    """
    REGISTER = SYMBOLIC.REGISTER_VARIABLE
    MEMORY = SYMBOLIC.MEMORY_VARIABLE


op_str = {AST_NODE.ANY: "any", AST_NODE.ASSERT: "assert", AST_NODE.BV: "bv", AST_NODE.BVADD: "+",
          AST_NODE.BVAND: "&", AST_NODE.BVASHR: "bvashr", AST_NODE.BVLSHR: "bvlshr", AST_NODE.BVMUL: "*",
          AST_NODE.BVNAND: "bvnand", AST_NODE.BVNEG: "-", AST_NODE.BVNOR: "bvnor", AST_NODE.BVNOT: "not",
          AST_NODE.BVOR: "|", AST_NODE.BVROL: "bvrol", AST_NODE.BVROR: "bvror", AST_NODE.BVSDIV: "bvsdiv",
          AST_NODE.BVSGE: ">=s", AST_NODE.BVSGT: ">s", AST_NODE.BVSHL: "<<", AST_NODE.BVSLE: "<=s",
          AST_NODE.BVSLT: "<s", AST_NODE.BVSMOD: "bvsmod", AST_NODE.BVSREM: "bvsrem", AST_NODE.BVSUB: "-",
          AST_NODE.BVUDIV: "bvudiv", AST_NODE.BVUGE: ">=u", AST_NODE.BVUGT: ">u", AST_NODE.BVULE: "<=u",
          AST_NODE.BVULT: "<u", AST_NODE.BVUREM: "bvurem", AST_NODE.BVXNOR: "bvxnor", AST_NODE.BVXOR: "^",
          AST_NODE.COMPOUND: "compound", AST_NODE.CONCAT: "concat", AST_NODE.DECLARE: "declare", AST_NODE.DISTINCT: "distinct",
          AST_NODE.EQUAL: "=", AST_NODE.EXTRACT: "extract", AST_NODE.FORALL: "forall", AST_NODE.IFF: "iff",
          AST_NODE.INTEGER: "integer", AST_NODE.ITE: "ite", AST_NODE.LAND: "land", AST_NODE.LET: "let",
          AST_NODE.LNOT: "lnot", AST_NODE.LOR: "lor", AST_NODE.REFERENCE: "reference", AST_NODE.STRING: "string",
          AST_NODE.SX: "sx", AST_NODE.VARIABLE: "variable", AST_NODE.ZX: "zx"}


class TritonAst:
    """
    Wrapping class on top of Triton AstNode objects. This is the main entity manipulated
    throughout the synthesis process. It provides many utility fonctions on these ASTs
    like :attr:`TritonAst.node_count` holding the number of node of the AST, or
    :meth:`TritonAst.reassembly` that allows reassembling the AST into assembly.


    """

    def __init__(self, ctx: TritonContext, node: AstNode, node_c: int, depth: int, vars: SymVarMap, children: List['TritonAst']) -> 'TritonAst':
        """
        Instanciate a TritonAst with some precomputed fields given in parameters.

        :param ctx: Triton context
        :type ctx: `TritonContext <https://triton.quarkslab.com/documentation/doxygen/py_TritonContext_page.html>`_
        :param node: Triton AstNode to wrap
        :type node: :py:obj:`qsynthesis.types.AstNode`
        :param node_c: Number of nodes contained in the expression
        :type node_c: int
        :param depth: Depth of the expression (depth of the AST)
        :type depth: int
        :param vars: Variables contained in this expression
        :type vars: Dict[str, :py:obj:`qsynthesis.types.SymbolicVariable`]
        :param children: List of children as TritonAst instances
        :type children: List[TritonAst]

        .. warning:: This class is not meant to be instanciated directly. It must be instanciated
             trough the :meth:`~TritonAst.make_ast` method.
        """
        self.ctx = ctx
        self.ast = self.ctx.getAstContext()
        self.expr = node  # It needs to be unrolled !!!
        self.size = self.expr.getBitvectorSize()
        self._symvars = vars  # SymVarName -> SymbolicVariable object  e.g: {'SymVar_1': rdi, 'SymVar_2': rsi}
        self._mapping = {chr(x[0]): x[1] for x in zip(range(97, 127), self._symvars.values())}  # {'a': SymVar, 'b': SymVar}
        self._parents = set()

        self._node_count = node_c
        self._depth = depth
        self._children = children

    @property
    def parents(self) -> List['TritonAst']:
        """
        Return the list of parents of a given AST. An AST is meant to have only
        ONE parent but Triton share common expression with multiple parents (wihtin
        the same expression)

        :rtype: List[TritonAst]
        """
        return list(self._parents)

    @property
    def mapping(self) -> Dict[Char, SymbolicVariable]:
        """
        Mapping a placeholder character ('a', 'b', 'c' ..) to all the SymbolicVariable
        of the object.

        :rtype: Dict[:py:obj:`qsynthesis.types.Char`, :py:obj:`SymbolicVariable`]
        """
        return self._mapping

    @mapping.setter
    def mapping(self, value: Dict[Char, SymbolicVariable]) -> None:
        """
        Set the given mapping of placeholder to their SymbolicVariable in the object.
        """
        self._mapping = value

    @property
    def sub_map(self) -> Dict[Char, AstNode]:
        """
        Similar to mapping but map a placeholder character ('a', 'b', 'c' ..) to
        the AstNode counterpart of SymbolicVariables.

        :rtype: Dict[:py:obj:`qsynthesis.types.Char`, :py:obj:`qsynthesis.types.AstNode`]
        """
        return {k: self.ast.variable(v) for k, v in self.mapping.items()}

    @property
    def type(self) -> AstType:
        """
        Returns the type of current AstNode object. The
        type is identical to the AST_NODE enum of Triton.

        :rtype: :py:obj:`qsynthesis.types.AstType`
        """
        return AstType(self.expr.getType())

    @property
    def hash(self) -> int:
        """
        Returns the Triton hash of the AstNode. This hash is meant to be unique
        for all AstNode, but is also meant to be similar to commutative expressions.

        :rtype: int
        """
        return self.expr.getHash()

    @property
    def ptr_id(self) -> int:
        """
        Returns the hash of the AstNode object. This attribute is meant to differentiate
        to different python object have the exact same AST structure.

        :rtype: int
        """
        return hash(self.expr)

    def is_constant_expr(self) -> bool:
        """Returns whether the AST expression is constant or not (namely does not
        have any symbolic variables in it)."""
        return len(self.symvars) == 0

    def is_variable(self) -> bool:
        """Returns of the TritonAst is a variable node."""
        return self.type == AstType.VARIABLE

    def is_constant(self) -> bool:
        """Returns True if the type of the object is Bitvector (namely constant)"""
        return self.type == AstType.BV

    @property
    def variable_id(self) -> int:
        """
        Get the Triton unique id for a variable.

        :raises: KeyError
        """
        if self.is_variable():
            return self.expr.getSymbolicVariable().getId()
        else:
            raise KeyError("not a variable")

    @property
    def var_num(self) -> int:
        """Returns the number of different symbolic variables of the expression

        :rtype: int
        """
        return len(self._symvars)

    @property
    def pp_str(self) -> str:
        """Hacky function that strips masks used in the AST_REPRESENTATION.PYTHON
        of Triton.

        :rtype: str
        """
        return str(self.expr).replace(" & 0xFFFFFFFFFFFFFFFF", "").replace(" & 0xffffffffffffffff", "")


    def visit_expr(self) -> Generator['TritonAst', None, None]:
        """ Pre-Order visit of all the sub-AstNode"""
        def rec(e):
            yield e
            for c in e.get_children():
                yield from rec(c)
        yield from rec(self)

    @property
    def symvars(self) -> List[SymbolicVariable]:
        """Returns the list of SymbolicVariable object of the current object

        :rtype: List[:py:obj:`qsynthesis.types.SymbolicVariable`]
        """
        return list(self._symvars.values())

    @staticmethod
    def symvar_type(v: SymbolicVariable) -> SymVarType:
        """
        Static method returning the type of a given symbolic variable object

        :param v: symbolic variable object
        :type v: :py:obj:`qsynthesis.types.SymbolicVariable`
        :return: Type of the symbolic variables
        :rtype: :py:obj:`qsynthesis.types.SymVarType`
        """
        return SymVarType(v.getType())

    def get_children(self) -> List['TritonAst']:
        """Return the list of children TritonAst"""
        return self._children

    def has_children(self) -> bool:
        """True whether the object has children or not"""
        return self.get_children() != []

    def is_root(self) -> bool:
        """
        Return True whether the object is a root node (namely does not have
        any parents).
        """
        return not(bool(self.parents))

    def is_leaf(self) -> bool:
        """True if the AST has no children"""
        return self.get_children() == []

    @property
    def node_count(self) -> int:
        """Pre-computed O(1) count of the number of node contained in this AST.

        :rtype: int
        """
        return self._node_count

    @property
    def depth(self) -> int:
        """Pre-computed depth of the AST *(complexity O(1))*

        :rtype: int
        """
        return self._depth

    @property
    def symbol(self) -> str:
        """Returns the symbol of the current AstNode, operator if binary expression
        variable name if variable or constant value if constant.

        :rtype: str
        """
        t = self.type
        if t in [AstType.BV, AstType.VARIABLE]:
            return str(self.expr)
        elif t == AstType.INTEGER:
            return str(self.expr.getInteger())
        else:
            return op_str[t]

    def mk_constant(self, v: int, size: int) -> 'TritonAst':
        """Create a new constant as a TritonAst (holding a Triton bv object)."""
        return TritonAst(self.ctx, self.ast.bv(v, size), 1, 1, {}, [])

    def mk_variable(self, alias: str, size: int) -> 'TritonAst':
        """
        Create a new variable node as a TritonAst (holding a Triton variable object).
        The variable is created in the TritonContext of the current object.
        """
        s = self.ctx.newSymbolicVariable(size, alias)
        s.setAlias(alias)
        ast_s = self.ast.variable(s)
        return TritonAst(self.ctx, ast_s, 1, 1, {s.getName(): s}, [])

    def normalized_str_to_ast(self, s: str) -> 'TritonAst':
        """
        Evaluate expression like "a + b - 1" creating a triton AST expression out
        of it. All variables have to be present in the AST.

        :param s: expression to evaluate
        :return: Triton AST node of the expression

        .. warning:: the str expr must be obtained through the eval_oracle of the
                     exact same TritonAst (otherwise names would be shuffled)
        """
        try:
            e = eval(s, self.sub_map)
        except NameError as e:
            logger.warning(f"Expression {s} evaluation failed {self.sub_map}")
            raise e
        except TypeError as exc:
            logger.error(f"Type error when evaluating: {s}")
            # Try mangling expressions
            nast = self._try_mangle_ast(s, exc)
            if nast:
                return nast
            else:
                raise exc
        ast = self.make_ast(self.ctx, e)
        return ast

    def _try_mangle_ast(self, s: str, e: Exception) -> Optional['TritonAst']:
        """
        Hacky function to mangle expression like 'a**2' generated by sympy.
        Triton AST and SMT does not support pow operators thus they have to
        be unrolled. This function tries to do it and if succeed try to generate
        an AST out of it.

        :param s: expression to evaluate
        :return: Triton AST node of the expression
        """
        logger.warning(f"Try mangling {s}: {str(e)}")
        import re
        mod = False
        s2 = s
        if str(e) == "unsupported operand type(s) for ** or pow(): 'AstNode' and 'int'":
            for c, v in re.findall("([^\(*]+)\*\*(\d+)", s):
                sub_c = c
                trail = ""
                if c.count("(") != c.count(")") and c.endswith(")"):
                    sub_c = sub_c[:-1]
                    trail = ")"
                # logger.info(f"Found instance {c} with {v}")
                mod = True
                s2 = re.sub(re.escape(c+"**"+v), '*'.join([f"({sub_c})"]*int(v))+trail, s2)
            logger.warning(f"Mangle expr {s} to {s2}")
        if mod and s2 != s:
            return self.normalized_str_to_ast(s2)
        else:
            return None

    def to_normalized_str(self) -> str:
        """
        Normalize the AST (replace variables by placeholder 'a', 'b' ..) and return
        it as a string.
        """
        back = {}
        for name, ast_v in self.sub_map.items():  # Substitute aliases with the normalized names
            sym_v = ast_v.getSymbolicVariable()
            back[name] = sym_v.getAlias()
            sym_v.setAlias(name)
        final_s = self.pp_str  # FIXME: Make a proper iteration of the expression to generate something compliant with lookup tables
        for name, ast_v in self.sub_map.items():  # Restore true aliases
            sym_v = ast_v.getSymbolicVariable()
            sym_v.setAlias(back[name])
        return final_s

    def update_all(self) -> None:
        """
        Update all children recursively of the current object. Fields being updated
        are node_count, depth and symvars. This might be used when some of the AST
        has been rewritten. All pre-computed values are then 'dirty' and have to be
        updated.
        """
        def rec(a):
            chs = [rec(x) for x in a.get_children()]
            if chs:  # one of their child might have been updated so update
                a.update()
                return a
            else:
                return a
        rec(self)

    def update(self) -> None:
        """
        Update the current AST node, with information of its directly children.
        Information of children are thus considered genuine.
        """
        chs = self.get_children()
        if chs:
            self._symvars = reduce(lambda acc, x: dict(acc, **x._symvars), chs, {})
            self._depth = max((x.depth for x in chs), default=0)+1
            self._node_count = sum(x.node_count for x in chs)+1

    def update_parents(self, recursive: bool = True) -> None:
        """
        Update the parent of the current TritonAst. `recursive` indicates
        if it has to be performed recrusively. If so the complexity of the
        operation O(depth).
        :param recursive: whether to apply it recursively on parents
        """
        for p in self.parents:
            p.update()
        if recursive:
            for p in self.parents:
                p.update_parents(recursive=recursive)

    def set_child(self, i: int, ast: 'TritonAst', update_node: bool = True, update_parents: bool = False) -> None:
        """
        Replace the ith child of the current TritonAst with new given ast object. Optional
        parameters indicates if inner fields of the object and its parent have to be updated.

        :param i: index of the child to replace
        :param ast: TritonAst to be used as replacement of the child
        :param update_node: whether to update internal field of the current node (node_count, depth, symvars)
        :param update_parents: whether to update parents
        """
        if isinstance(i, TritonAst):
            i = {c: i for i, c in enumerate(self.get_children())}[i]  # retrieve number from instance
        if update_node:
            if not self.expr.setChild(i, ast.expr):
                assert False
        self._children[i] = ast  # implicitely unlink previous ast and replace it by the new
        ast._parents.add(self)   # add self as parent of the new ast
        self.update()            # update its own fields
        if update_parents:
            self.update_parents()

    def replace_self(self, repl: 'TritonAst', update_parents: bool = True) -> None:
        """
        Replace the current object by the given TritonAst. This function thus replace
        parents by replace the child that correspond to the current object by the replacement.
        :param repl: TritonAst used to replace the current object
        :param update_parents: whether to update parents or not
        """
        if len(self.parents):
            logger.debug("replace self while multiple parents !")
        is_first = True
        for p in self.parents:
            p.set_child(self, repl, update_node=is_first, update_parents=update_parents)
            self._parents.remove(p)  # remove its own parent
            # is_first = False

    @staticmethod
    def make_ast(ctx: TritonContext, exp: AstNode) -> 'TritonAst':
        """
        Main staticmethod meant to create all TritonAst object. This method iterates
        all the given expression ``expr`` recursively to create TritonAst's
        all the way down and pre-computing along the way the important fields like
        node_count, depth and symvars.

        :param ctx: Triton context on which to work on
        :param exp: AstNode object to iterate
        :returns: TritonAst instance wrapping the exp object
        """
        ptr_map = {}  # hash -> TritonAst

        def rec(e):
            h = hash(e)
            if h in ptr_map:  # if the pointer as already been seen
                return ptr_map[h]  # return it early
            typ = e.getType()
            if typ == AST_NODE.REFERENCE:
                return rec(e.getSymbolicExpression().getAst())
            elif typ == AST_NODE.BV:
                t = TritonAst(ctx, e, 1, 1, {}, [])
            elif typ == AST_NODE.INTEGER:
                t = TritonAst(ctx, e, 1, 1, {}, [])
            elif typ == AST_NODE.VARIABLE:
                symvar = e.getSymbolicVariable()
                name = symvar.getName()
                t = TritonAst(ctx, e, 1, 1, {name: symvar}, [])
            else:
                chs, symvs, node_c = [], {}, 1
                for ast_child in e.getChildren():
                    c = rec(ast_child)
                    symvs.update(c._symvars)
                    node_c += c.node_count
                    chs.append(c)
                depth = max(c.depth for c in chs)+1
                t = TritonAst(ctx, e, node_c, depth, symvs, chs)
                for c in chs:
                    c._parents.add(t)
            ptr_map[hash(e)] = t
            return t
        return rec(exp)

    def duplicate(self) -> 'TritonAst':
        """
        Create a new distinct instance of TritonAst
        """
        new_expr = self.ast.duplicate(self.expr)
        return TritonAst.make_ast(self.ctx, new_expr)

    def random_sampling(self, n: int) -> IOVector:
        """
        Generates a random list of I/O samples pair.

        :param n: number of samples to generate
        :return: a list of n (inputs, output) tuples
        :rtype: :py:obj:`qsynthesis.types.IOVector`
        """
        samples = []

        for _ in range(n):
            inputs = {k: random.getrandbits(v.getBitvectorSize()) for k, v in self.sub_map.items()}
            output = self.eval_oracle(inputs)
            samples.append((inputs, output))
        return samples

    def eval_oracle(self, inp: Input) -> Output:
        """
        Oracle corresponding to the wrapped AST. It takes a valuation for all its
        symbolic variables and as an oracle returns the associated output.

        :param inp: a mapping of variable to a given input value which will be used
                    as concrete values for the symbolic variables in the wrapped AST
        :type inp: :py:obj:`qsynthesis.types.Input`
        :return: The result computed by means of evaluating the AST
        :rtype: :py:obj:`qsynthesis.types.Output`
        """
        for v_name, symvar in self.mapping.items():
            self.ctx.setConcreteVariableValue(symvar, inp[v_name])
        return self.expr.evaluate()

    def compare_behavior(self, ast: TritonAst, inps: List[Input]) -> int:
        """
        Compare the current expression with the one given in parameter wrt theirs
        behavior on the given set of inputs. The comparison returns -1 if not applicable
        (as involving different variables, 0 if different and 1 if equal.

        :param ast: other ast to compare against
        :param inps: Set of inputs to use for evaluation
        :return: -1 if not applicable, 0 if different and 1 if equal
        :rtype: int
        """
        if set(x.getAlias() for x in self.mapping.values()).symmetric_difference(set(x.getAlias() for x in ast.mapping.values())):
            return -1  # Some variables are different
        backup = self.mapping
        print([self.eval_oracle(i) for i in inps])
        rev = {v.getAlias(): k for k, v in ast.mapping.items()}  # Map name -> pld   e.g: {'rsi': 'a', 'rdi': 'b'}
        self.mapping = {rev[vs.getAlias()]: vs for pld, vs in self.mapping.items()}  # keep own vars (vs) but remap on other ast identifier
        print(backup, self.mapping)
        tmp1 = [self.eval_oracle(i) for i in inps]
        tmp2 = [ast.eval_oracle(i) for i in inps]
        print(tmp1, "\n", tmp2)
        res = tmp1 == tmp2
        self.mapping = backup  # Restore backup submap
        return res

    def to_z3(self) -> 'z3.z3.ExprRef':
        """Returns the Z3 expression associated with the Triton AST expression"""
        return self.ast.tritonToZ3(self.expr)

    @staticmethod
    def from_z3(ctx: TritonContext, expr: 'z3.z3.ExprRef') -> 'TritonAst':
        """
        Create a TritonAst out of a Z3 expressions

        :param ctx: Triton Context in which to create the expression
        :param expr: Z3 expression
        :return: TritonAst

        .. warning:: This function is mostly untested !
        """
        astctx = ctx.getAstContext()
        ast = astctx.z3ToTriton(expr)
        return TritonAst.make_ast(ctx, ast)

    def is_semantically_equal(self, other: 'TritonAst') -> bool:
        """
        Allows checking if the current AST is semantically equal to the one provided.

        :param other: TritonAst on which to test against
        :returns: bool -- True if both ASTs are semantically equals
        """
        cst = self.ast.distinct(self.expr, other.expr)
        return not(self.ctx.isSat(cst))

    @staticmethod
    def dyn_node_count(expr: AstNode) -> int:
        """
        Returns the effective count of node of the expression by iterating the
        AstNode object recursively. The complexity is O(N) with N the number of node.

        :param expr: AstNode to iterate
        :type expr: :py:obj:`qsynthesis.types.AstNode`
        :returns: Number of nodes in the AST.

        .. note:: The way of counting nodes is different from the number of nodes of
                  Triton for which bitvector values are composed of 3 nodes. We count
                  them as one.
        """
        def rec(e):
            typ = e.getType()
            if typ == AST_NODE.REFERENCE:
                return rec(e.getSymbolicExpression().getAst())
            elif typ == AST_NODE.BV:
                return 1  # prevent counting BV childs as nodes
            else:
                return 1+sum(map(rec, e.getChildren()))
        return rec(expr)

    @staticmethod
    def dyn_depth(expr: AstNode) -> int:
        """
        Returns the effective depth of the node of the expression by iterating the
        AstNode object recursively. The complexity is O(N) with N the depth of the AST.

        :param expr: AstNode to iterate
        :type expr: :py:obj:`qsynthesis.types.AstNode`
        :returns: AST depth
        """
        def rec(e):
            typ = e.getType()
            if typ == AST_NODE.REFERENCE:
                return rec(e.getSymbolicExpression().getAst())
            elif typ == AST_NODE.BV:
                return 1  # prevent couting BV childs as nodes
            else:
                return 1+max(map(rec, e.getChildren()), default=0)
        return rec(expr)

    @staticmethod
    def _visit_replacement(ast: 'TritonAst') -> Generator[Union['TritonAst', Tuple['TritonAst', bool]], 'TritonAst', None]:
        rep = yield ast  # First (Top-Down yield)

        if rep is not None:  # We should replace this expr
            yield rep, True  # Final (yield)
        else:
            rep_took_place = False
            for i, c in enumerate(ast.get_children()):  # Iterate (and simplify) all childrens
                g = TritonAst._visit_replacement(c)
                recv = None
                while 1:
                    it = g.send(recv)  # forward the new expression to the sub-generator
                    if isinstance(it, tuple):  # is a final yield
                        if it[1]:  # A replacement took place
                            rep_took_place = True
                            ast.set_child(i, it[0])
                        break
                    else:
                        recv = yield it  # re-yield it above (and receive a new expression)
            yield ast, rep_took_place

    def _inplace_replace(self, other: 'TritonAst') -> None:
        logger.debug("Inplace replace")
        self.ctx, self.ast = other.ctx, other.ast
        self.expr = other.expr
        self.size = other.size
        self._symvars = other._symvars
        self._parents = other._parents
        self._node_count, self._depth = other._node_count, other._depth
        self._children = other._children

    def visit_replacement(self, update: bool = True) -> Generator['TritonAst', 'TritonAst', None]:
        """
        Triton AST expression replacement visitor in a Top-Down manner. It yields
        every sub-expression and replace it with the expression received throught
        the send mechanism.

        :param update: whether to update each node after having been replaced
        :return: generator of TritonAst, which for each AST yielded wait to
                 receive either None meaning the it should not be replaced or a new
                 TritonAst to be put in replacement.
        """
        v = self._visit_replacement(self)
        new_expr_to_send = None
        while True:
            cur_expr = v.send(new_expr_to_send)
            if isinstance(cur_expr, tuple):
                if update:
                    self.update_all()
                break
            else:
                new_expr_to_send = yield cur_expr
                if new_expr_to_send is not None and cur_expr == self:  # Changing root node so fusion fields and return
                    self._inplace_replace(new_expr_to_send)
                    return

    def reassemble(self, dst_reg: str, target_arch: Optional[str] = None) -> bytes:
        """
        Reassemble the TritonAst in assembly. ``dst_reg`` is the destination register of the
        result of the computation of the AST. Parameter ``target_arch`` is either a QtraceDB
        architecture object or the string identifier of the architecture as defined by the
        LLVM architecture triple: https://llvm.org/doxygen/classllvm_1_1Triple.html#a547abd13f7a3c063aa72c8192a868154
        If no architecture is provided, use the same than the one that the AST.

        :param dst_reg: destination register as lowercase string
        :param target_arch: target architecture in which to reassemble the AST
        :returns: bytes of the AST reassembled in the given architecture
        :raises: ReassemblyError

        .. warning:: This method requires the ``arybo`` library that can be installed with
           (pip3 install arybo).
        """
        def my_asm_binary(arybo_expr, dst_regs, inps, target):
            # FIXME: ``asm_binary`` of arybo is broken as it used the function
            # ObjectFileRef of llvmlite which is itself broken. Up until a fix
            # is pushed I have to reimplement asm_binary and retrieve bytes with
            # lief. cf: https://github.com/numba/llvmlite/issues/632
            mod = asm_module(arybo_expr, dst_regs, inps, target)
            M = llvm.parse_assembly(str(mod))
            M.verify()
            target = llvm_get_target(target)
            machine = target.create_target_machine()
            obj_bin = machine.emit_object(M)
            p = lief.parse(obj_bin)
            return bytes(p.get_section('.text').content)
        try:
            import lief
            from arybo.tools.triton_ import tritonast2arybo
            from arybo.lib.exprs_asm import asm_binary, asm_module, llvm, llvm_get_target
            if all(TritonAst.symvar_type(x) == SymVarType.REGISTER for x in self.symvars):
                if target_arch is None:
                    m = {ARCH.X86: "x86", ARCH.X86_64: "x86_64", ARCH.ARM32: "arm", ARCH.AARCH64: "aarch64"}
                    arch_name = m[self.ctx.getArchitecture()]
                else:
                    arch_name = target_arch.lower()
                arybo_expr = tritonast2arybo(self.expr, use_exprs=True, use_esf=False)
                inps = {x.getName(): (x.getAlias(), x.getBitSize()) for x in self.symvars}
                return my_asm_binary(arybo_expr, (dst_reg, self.size), inps, f"{arch_name}-unknown-unknwon")
            else:
                raise ReassemblyError("Can only reassemble if variable are registers (at the moment)")
        except ImportError as e:
            raise ReassemblyError(f"Cannot import arybo, while it is required (pip3 install arybo): {e}")
        except NameError:
            raise ReassemblyError(f"Invalid target architecture '{target_arch}' provided")
        except Exception as e:
            raise ReassemblyError(f"Something went wrong during reassembly: {e}")

    def make_graph(self) -> 'Graph':
        """
        Generate a graph object representing the AST
        as a graph_tool object.

        .. warning:: This method requires the ``graph_tool`` python library that can be installed
           by following `https://git.skewed.de/count0/graph-tool/-/wikis/installation-instructions`
        """
        from graph_tool import Graph
        graph = Graph(directed=True)

        graph.vp['sym'] = graph.new_vertex_property('string')
        graph.vp['vmap'] = graph.new_vertex_property('int64_t')
        n = graph.add_vertex(1)

        worklist = [(None, (self, n))]
        while worklist:
            na, (b, nb) = worklist.pop(0)

            graph.vp['sym'][nb] = b.symbol
            graph.vp['vmap'][nb] = id(b)

            if na is not None:
                graph.add_edge(na, nb)

            for c in b.get_children():
                nc = graph.add_vertex(1)
                worklist.append((nb, (c, nc)))
        return graph



# built-in modules
from typing import List

# qsynthesis modules
from qsynthesis.plugin.dependencies import ida_graph, Instr
from qsynthesis.tritonast import TritonAst


class AstViewer(ida_graph.GraphViewer):
    """
    GraphViewer to show AST Triton expressions
    """

    def __init__(self, title: str, ast: TritonAst):
        """
        Constructor.

        :param title: string title of the windows
        :param ast: TritonAst object to show
        """
        ida_graph.GraphViewer.__init__(self, title)
        self._ast = ast

    def OnRefresh(self) -> bool:
        """
        GraphViewer callback called whenever refreshing the view (so at least once)
        """
        self.Clear()
        self.draw()
        return True

    def OnGetText(self, ida_node_id: int) -> str:
        """
        GraphViewer callback whenever it needs to get the text associated to a node.
        Just return the string pre-computed in :meth:`AstViewer.draw`.

        :param ida_node_id: GraphViewer node id
        :return: text of the node
        """
        return self[ida_node_id]

    def Show(self) -> bool:
        """
        Called when showing the view
        """
        # TODO: Checking if its really required ?
        if not ida_graph.GraphViewer.Show(self):
            return False
        return True

    def draw(self) -> None:
        """
        Create the graph by iterating the AST and creating
        associated node in the GraphViewer object. The text
        of each node is their 'symbol' attribute.
        """
        n_id = self.AddNode(self._ast.symbol)
        worklist = [(n_id, self._ast)]

        while worklist:
            node_id, node = worklist.pop(0)

            for c in node.get_children():
                child_id = self.AddNode(c.symbol)
                self.AddEdge(node_id, child_id)
                worklist.append((child_id, c))


class BasicBlockViewer(ida_graph.GraphViewer):
    """
    Class to show handcrafted basic blocks with
    a given set of instructions.
    """
    def __init__(self, title: str, insts: List[Instr]):
        """
        Constructor. Takes window title and a list of instruction
        to show.
        """
        ida_graph.GraphViewer.__init__(self, title)
        self.insts = insts
        # TODO: Using ida_lines stuff to precompute 'colored' lines

    def OnRefresh(self) -> bool:
        """
         GraphViewer callback called whenever refreshing the view (so at least once)
         """
        self.Clear()
        self.draw()
        return True

    def OnGetText(self, ida_node_id: int) -> str:
        """
        GraphViewer callback whenever it needs to get the text associated to a node.
        Just return the string pre-computed in :meth:`BasicBlockViewer.draw`.

        :param ida_node_id: GraphViewer node id
        :return: text of the node
        """
        return self[ida_node_id]

    def Show(self):
        """ Called when showing the view """
        return False if not ida_graph.GraphViewer.Show(self) else True

    def draw(self) -> None:
        """ Add a single basic block corresponding to the concatenation of string instructions """
        self.AddNode("\n".join(str(x) for x in self.insts))



from unittest.mock import MagicMock

try:
    import ida_idaapi
    import ida_kernwin
    import ida_gdl
    import ida_funcs
    import ida_bytes
    import ida_idp
    import ida_nalt
    import ida_ua
    import ida_loader
    import ida_auto
    import ida_graph
    import ida_lines
    IDA_ENABLED = True
except ImportError:
    # If we cannot import ida Mock most of its API to works equally without
    class ida_kernwin:
        class PluginForm:
            pass
        action_handler_t = MagicMock()
    ida_idaapi = MagicMock()
    ida_gdl = MagicMock()
    ida_bytes = MagicMock()
    ida_idp = MagicMock()
    ida_nalt = MagicMock()
    ida_funcs = MagicMock()
    ida_ua = MagicMock()
    ida_loader = MagicMock()
    ida_auto = MagicMock()
    ida_graph = MagicMock()
    ida_lines = MagicMock()

    IDA_ENABLED = False


try:
    import qtraceida
    QTRACEIDA_ENABLED = True
except ImportError:
    QTRACEIDA_ENABLED = False


try:
    import qtracedb
    from qtracedb import DatabaseManager
    from qtracedb.trace import Trace, InstrCtx
    from qtracedb.archs.arch import Instr, Arch
    from qtracedb.manager import ArchsManager
    from qtracedb.archs.x86 import ArchX86, ArchX64
    from qtracedb.archs.arm import ArchARM
    from qtracedb.archs.arm64 import ArchARM64
    QTRACEDB_ENABLED = True
except ImportError as e:
    QTRACEDB_ENABLED = False
    from .arch import Arch, ArchX64, ArchARM, ArchARM64, ArchX86, Instr, ArchsManager
    # Set dummy vars
    DatabaseManager, Trace, InstrCtx = None, None, None


try:
    import triton
    TRITON_ENABLED = True
except ImportError:
    TRITON_ENABLED = False



# built-in libs
from __future__ import annotations
from typing import List, Set, Iterable, Union, Optional

# third-party libs
from triton import ARCH, CALLBACK, MODE, MemoryAccess, Instruction, AST_REPRESENTATION, OPERAND
from triton import TritonContext

# qsynthesis deps
from qsynthesis.tritonast import TritonAst
from qsynthesis.types import SymbolicExpression, Register, Addr, ByteSize, SymbolicVariable


class SimpleSymExec:
    """
    Helper class used to process a sequential bunch of instructions using Triton
    and mark as symbolic all memory and registers reads that occoured before a
    write to that same register or memory address
    """

    def __init__(self, arch: Union[ARCH, str]):
        """
        Initialize symbolic execution

        :param arch: Triton architecture identifier or string of it
        :type: `ARCH <https://triton.quarkslab.com/documentation/doxygen/py_ARCH_page.html>`_
        """
        arch = getattr(ARCH, arch.upper()) if isinstance(arch, str) else arch
        self.ctx = TritonContext(arch)
        self.ctx.setMode(MODE.ALIGNED_MEMORY, True)
        self.ctx.setMode(MODE.AST_OPTIMIZATIONS, True)
        self.ctx.setMode(MODE.CONSTANT_FOLDING, True)
        self.ctx.setMode(MODE.ONLY_ON_SYMBOLIZED, True)
        self.ctx.setAstRepresentationMode(AST_REPRESENTATION.PYTHON)

        self.mem_addr_seen = set()  # All addresses read / written during execution
        self.mem_symvars = []       # All memory symbolic variables
        self.reg_id_seen = set()    # Triton Id of registers written directly or initialized with a value
        self.reg_symvars = []       # All registers symbolic variables

        self._capturing = False

        # Add callbacks
        self.ctx.addCallback(CALLBACK.SET_CONCRETE_MEMORY_VALUE, self._mem_write_callback)
        self.ctx.addCallback(CALLBACK.GET_CONCRETE_MEMORY_VALUE, self._mem_read_callback)
        self.ctx.addCallback(CALLBACK.SET_CONCRETE_REGISTER_VALUE, self._reg_write_callback)
        self.ctx.addCallback(CALLBACK.GET_CONCRETE_REGISTER_VALUE, self._reg_read_callback)

        self.cur_inst = None
        self.inst_id = 0
        self._expr_id = 0
        self._inst_cbs = []

    @property
    def expr_id(self) -> int:
        """
        Within a single instruction, used to represent an expression id.
        Programmed as an auto-increment variable upon each read.

        :rtype: int
        """
        self._expr_id += 1
        return self._expr_id - 1

    @expr_id.setter
    def expr_id(self, value: int) -> None:
        """Set the given expr_id"""
        self._expr_id = value

    @property
    def arch(self) -> ARCH:
        """Return the triton architecture identifier of the current context

        :rtype: `ARCH <https://triton.quarkslab.com/documentation/doxygen/py_ARCH_page.html>`_
        """
        return self.ctx.getArchitecture()

    @property
    def flags_reg(self) -> Register:
        """
        Portable function the get the flag register accross all Triton
        supported architectures.

        :rtype: `Register <https://triton.quarkslab.com/documentation/doxygen/py_Register_page.html>`_
        """
        _mapper = {ARCH.X86: "eflags", ARCH.X86_64: "eflags", ARCH.ARM32: "cpsr", ARCH.AARCH64: "spsr"}  # Not really true for spsr
        return getattr(self.ctx.registers, _mapper[self.arch])

    @property
    def ins_ptr_reg(self) -> Register:
        """
        Portable function get get the instruction pointer register depending
        on the current architecture.

        :rtype: `Register <https://triton.quarkslab.com/documentation/doxygen/py_Register_page.html>`_
        """
        _mapper = {ARCH.X86: "eip", ARCH.X86_64: "rip", ARCH.ARM32: "pc", ARCH.AARCH64: "pc"}
        return getattr(self.ctx.registers, _mapper[self.arch])

    @property
    def current_address(self) -> Addr:
        """Return the address of the current instruction.

        :rtype: :py:obj:`qsynthesis.types.Addr`
        """
        return self.cur_inst.getAddress()

    def _turn_on(self) -> None:
        """Turn on capturing"""
        self._capturing = True

    def _turn_off(self) -> None:
        """Turn off capturing"""
        self._capturing = False

    def _mem_write_callback(self, _: TritonContext, ma: MemoryAccess, __: int) -> None:
        """
        Callback called by Triton upon each memory write. It is used to update a map
        of all addresses written.
        """
        if not self._capturing:
            return

        # Add bytes to set of written bytes
        self.mem_addr_seen |= self._memacc_to_all_addr(ma)

    def _mem_read_callback(self, ctx: TritonContext, ma: MemoryAccess) -> None:
        """
        Callback called by Triton upon each memory read. It is used to dynamically
        symbolized memory cells accessed that have not been seen before.
        """
        if not self._capturing:
            return

        # Retrieve all addresses of a given mem access
        addrs = self._memacc_to_all_addr(ma)

        # Get addrs which have not been written to and are not
        # yet been symbolized (eg. are not yet parameters)
        new_addrs = addrs - self.mem_addr_seen

        # Stop if we don't have any new address. This is also needed to stop
        # recursion since convertMemoryToSymbolicVariable will trigger again this callback
        if not new_addrs:
            return

        # Update mem_operands
        self.mem_addr_seen |= new_addrs

        # Symbolize all the memory cells that have not yet been seen
        # Coalesce adjacent bytes and create memory accesses' symvars
        for ma in self._coalesce_bytes_to_mas(new_addrs):
            symvar = self.symbolize_memory(ma)
            self.mem_symvars.append(symvar)

    def _reg_write_callback(self, ctx: TritonContext, reg: Register, _: int) -> None:
        """
        Callback called by Triton on each register write.
        """
        if not self._capturing:
            return

        # Add parent register to set of written registers)
        parent_reg = ctx.getParentRegister(reg)
        self.reg_id_seen.add(parent_reg.getId())

    def _reg_read_callback(self, ctx: TritonContext, reg: Register) -> None:
        """
        Callback called by Triton on each register read.
        """
        if not self._capturing:
            return

        parent_reg = ctx.getParentRegister(reg)
        parent_reg_id = parent_reg.getId()

        # Skip reg if it has already been written or is already an operand
        if parent_reg_id in self.reg_id_seen:
            return

        # Ignore flags and pc registers (and make sure we know it!)
        if parent_reg in [self.flags_reg, self.ins_ptr_reg]:
            return

        # Ignore registers which are immutables
        if not reg.isMutable():
            return

        # Symbolize the (full) register
        self.symbolize_register(parent_reg, 0)

    def get_register_ast(self, reg_name: Union[str, Register]) -> TritonAst:
        """
        Get the TritonAst associated with the given register. The register
        can either be a string or a triton register object.

        :param reg_name: register name or triton register
        :type reg_name: Union[str, `Register <https://triton.quarkslab.com/documentation/doxygen/py_Register_page.html>`_]
        :returns: the TritonAst associated to that register
        """
        reg = getattr(self.ctx.registers, reg_name.lower()) if isinstance(reg_name, str) else reg_name
        reg_se = self.ctx.getSymbolicRegister(reg)
        actx = self.ctx.getAstContext()
        if reg_se is None:
            e = actx.bv(self.ctx.getConcreteRegisterValue(reg), reg.getBitSize())
        else:
            e = actx.unroll(reg_se.getAst())
        return TritonAst.make_ast(self.ctx, e)

    def get_memory_ast(self, addr: Addr, size: ByteSize) -> TritonAst:
        """
        Get the TritonAst associated with the given address and size.

        :param addr: address of which to create the AST
        :type addr: :py:obj:`qsynthesis.types.Addr`
        :param size: Size of the read in memory (in bytes)
        :type size: :py:obj:`qsynthesis.types.ByteSize`
        :returns: the TritonAst of the memory content
        :rtype: TritonAst
        """
        ast = self.ctx.getMemoryAst(MemoryAccess(addr, size))
        actx = self.ctx.getAstContext()
        return TritonAst.make_ast(self.ctx, actx.unroll(ast))

    def get_operand_ast(self, op_num: int, inst: Optional[Instruction] = None) -> TritonAst:
        """
        Get the TritonAst of the of the ith operand. The instruction can be provided
        as an optional parameter. If not provided it takes the current instruction
        having been processed.

        :param op_num: operand number (starting at 0)
        :param inst: Triton Instruction of which to get the operand
        :type inst: Optional[`Instruction <https://triton.quarkslab.com/documentation/doxygen/py_Instruction_page.html>`_]
        :returns: the TritonAst of the operand
        :rtype: TritonAst
        """
        inst = self.cur_inst if inst is None else inst
        op = inst.getOperands()[op_num]
        actx = self.ctx.getAstContext()
        t = op.getType()
        if t == OPERAND.IMM:
            e = actx.bv(op.getValue(), op.getBitSize())
        elif t == OPERAND.REG:
            e = self.ctx.getRegisterAst(op)
        elif t == OPERAND.MEM:
            e = self.ctx.getMemoryAst(op)
        else:
            assert False
        return TritonAst.make_ast(self.ctx, actx.unroll(e))

    def get_register_symbolic_expression(self, reg_name: Union[str, Register]) -> SymbolicExpression:
        """
        Get the current SymbolicExpression (triton object) of the register given in parameter.

        :param reg_name: register name, or Register object
        :type reg_name: Union[str, `Register <https://triton.quarkslab.com/documentation/doxygen/py_Register_page.html>`_]
        :returns: current symbolic expression of the register
        :rtype: `SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_
        """
        reg = getattr(self.ctx.registers, reg_name.lower()) if isinstance(reg_name, str) else reg_name
        return self.ctx.getSymbolicRegister(reg)

    def get_memory_symbolic_expression(self, addr: Addr, size: ByteSize) -> SymbolicExpression:
        """
        Get the current SymbolicExpression of the given address. As no symbolic expression are
        assigned to single memory bytes. The function creates a new symbolic expression representing
        the addr+size expression.

        :param addr: address in memory
        :type addr: :py:obj:`qsynthesis.types.Addr`
        :param size: size in bytes of the memory read
        :type ByteSize: :py:obj:`qsynthesis.types.ByteSize`
        :returns: symbolic expression representing the memory content value
        :rtype: `SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_
        """
        ast = self.ctx.getMemoryAst(MemoryAccess(addr, size))
        sym = self.ctx.newSymbolicExpression(ast)
        sym.setComment(self._fmt_comment())
        return sym

    def get_operand_symbolic_expression(self, op_num: int) -> SymbolicExpression:
        """
        Get the symbolic expression of the ith operand of the current instruction being
        processed by the symbolic executor. Depending on the type of the operand
        calls :meth:`SimpleSymExec.get_register_symbolic_expression`, or
        :meth:`SimpleSymExec.get_memory_symbolic_expression`. For constants create
        a new symbolic expression.

        :param op_num: operand number
        :type op_num: int
        :returns: the symbolic expression of the operand
        :rtype: `SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_
        """
        op = self.cur_inst.getOperands()[op_num]
        t = op.getType()
        if t == OPERAND.IMM:
            ast = self.ctx.getAstContext().bv(op.getValue(), op.getBitSize())
            sym = self.ctx.newSymbolicExpression(ast)
            sym.setComment(self._fmt_comment())
        elif t == OPERAND.REG:
            sym = self.get_register_symbolic_expression(op)
        elif t == OPERAND.MEM:
            sym = self.get_memory_symbolic_expression(op.getAddress(), op.getSize())
        else:
            assert False
        return sym

    def symbolize_register(self, reg: Register, value: int) -> SymbolicVariable:
        """
        Symbolize the given register with the associated concrete value.

        :param reg: Register to symbolize
        :type reg: `Register <https://triton.quarkslab.com/documentation/doxygen/py_Register_page.html>`_
        :param value: Concrete value to assign the register (required for soundness)
        :type value: int
        :returns: the symbolic variable created for the register
        :rtype: `SymbolicVariable <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicVariable_page.html>`_
        """
        self.reg_id_seen.add(reg.getId())
        symvar = self.ctx.symbolizeRegister(reg, reg.getName())
        #symvar.setComment(comment)

        # Set comment on the register reference
        sreg = self.ctx.getSymbolicRegister(reg)
        sreg.setComment(self._fmt_comment())

        # We also set the symbolic var to the actual value of the register
        self.ctx.setConcreteVariableValue(symvar, value)
        self.reg_symvars.append(symvar)
        return symvar

    def symbolize_memory(self, mem: MemoryAccess) -> SymbolicVariable:
        """
        Symbolize the given triton MemoryAccess. Into a new SymbolicVariable.

        :param mem: memory access to symbolize
        :type mem: `MemAccess <https://triton.quarkslab.com/documentation/doxygen/py_MemoryAccess_page.html>`_
        :returns: symbolic variable representing the content
        :rtype: `SymbolicVariable <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicVariable_page.html>`_
        """
        # The issue here is that we need to assign a specific comment
        # on symbolic expression to attach them to a specific instruction
        # and being able to perform slicing afterward. But there is no
        # direct manner to get symbolic expressions created by the
        # symbolizeMemory

        # symbolize the MemAccess and obtain a SymbolicVariable object
        alias = f"mem_{mem.getAddress():#x}_{mem.getSize()}_{self.inst_id}"
        symvar = self.ctx.symbolizeMemory(mem, alias)
        symvar.setComment(self._fmt_comment())

        # Iterate each address of the MemAccess to obtain the SymbolicExpression
        # of the address. Put for each of them the right comment
        addr = mem.getAddress()
        end = addr + mem.getSize()
        cur_mem_exp = None
        while addr < end:
            cur_mem_exp = self.ctx.getSymbolicMemory(addr)
            cur_mem_exp.setComment(self._fmt_comment())
            addr += 1

        # Each SymbolicExpression have a uniq Id. The way they are created
        # (I heuristically know that the one before the last cur_mem_exp)
        # is the expression concating all of them thus also add a comment on it
        var_exp = self.ctx.getSymbolicExpression(cur_mem_exp.getId()-1)
        var_exp.setComment(self._fmt_comment())
        return symvar

    def initialize_register(self, reg: Union[str, Register], value: int) -> None:
        """
        Initialize a register by giving it an initial value. Register can
        either be a string or a Register object.

        :param reg: reg name string or register object
        :type reg: Union[str, `Register <https://triton.quarkslab.com/documentation/doxygen/py_Register_page.html>`_]
        :param value: integer value of the register
        """
        reg = getattr(self.ctx.registers, reg.lower()) if isinstance(reg, str) else reg
        self.reg_id_seen.add(reg.getId())
        self.ctx.setConcreteRegisterValue(reg, value)

    def disassemble(self, opcode: bytes, addr: Optional[Addr] = None) -> Instruction:
        """
        Disassemble a given opcode using Triton. Returns a Triton Instruction object.
        The Instruction is not been symbolically executed but the internal `cur_inst`
        attribute is set. The address is optional.
        If not provided the current instruction pointer in the internal context is used.

        :param opcode: bytes of the instruction
        :type opcode: bytes
        :param addr: address of the instruction
        :type addr: Optional[:py:obj:`qsynthesis.types.Addr`]
        :returns: Triton Instruction object
        :rtype: `Instruction <https://triton.quarkslab.com/documentation/doxygen/py_Instruction_page.html>`_
        """
        inst = Instruction(addr, opcode[:16]) if addr is not None else Instruction(opcode[:16])
        self.ctx.disassembly(inst)
        self.cur_inst = inst  # Set it if require to query get_operand_symbolic_expression
        return inst

    def execute_blob(self, data: bytes, addr: Addr) -> bool:
        """
        Execute instructions of the blob while the program counter
        remains in the blob. Consume the whole data and execute it
        symbolically.

        .. warning: This method writes payload into triton context.

        :param data: bytes of instructions to execute
        :type data: bytes
        :param addr:  address of the first instruction
        :type addr: :py:obj:`qsynthesis.types.Addr`
        :returns: True if execution of all instructions succeeded
        :rtype: bool
        """
        self.ctx.setConcreteMemoryAreaValue(addr, data)  # set concrete data in memory
        self.ctx.setConcreteRegisterValue(self.ins_ptr_reg, addr) # set program counter value
        pc = addr

        while addr <= pc < (addr + len(data)):  # while we remain in the blob
            opcode = self.ctx.getConcreteMemoryAreaValue(pc, 16) 
            if not self.execute(opcode, pc):
                return False
            pc =  self.ctx.getConcreteRegisterValue(self.ins_ptr_reg)
            
        return True

    def execute_basic_block(self, data: bytes, addr: Optional[Addr] = None) -> bool:
        """
        Execute a whole bunch of bytes as instructions. Consume the whole
        data and execute it symbolically. No payload is written in the triton
        context. The basic block is executed 'out of thin air'.

        :param data: bytes of instructions to execute
        :type data: bytes
        :param addr: optional address of the first instruction
        :type addr: Optional[:py:obj:`qsynthesis.types.Addr`]
        :returns: True if execution of all instructions succeeded
        :rtype: bool
        """
        blob = data[:]
        while blob:
            i = self.disassemble(blob, addr)
            addr = None  # reset to None if it was provided
            if not self.execute_instruction(i):
                return False
            blob = blob[i.getSize():]
        return True

    def execute(self, opcode: bytes, addr: Optional[Addr] = None) -> bool:
        """
        Symbolically execute the given opcode at the given optional address.

        :param opcode: bytes of the instruction
        :type opcode: bytes
        :param addr: optional address of the instruction
        :type addr: Optional[:py:obj:`qsynthesis.types.Addr`]
        :returns: True if the instruction has sucessfully been processed
        :rtype: bool
        """
        inst = Instruction(addr, opcode) if addr is not None else Instruction(opcode)
        return self.execute_instruction(inst)

    def execute_instruction(self, instr: Instruction) -> bool:
        """
        Symbolically execute the given triton Instruction already instanciated.

        :param instr: Triton Instruction
        :type instr: `Instruction <https://triton.quarkslab.com/documentation/doxygen/py_Instruction_page.html>`_
        :returns: True if the processing has successfully been performed
        :rtype: bool
        """
        # Update object values
        self.cur_inst = instr
        self.inst_id += 1
        self.expr_id = 0

        # Process instruction
        self._turn_on()

        # Call any instruction callback if any
        if self._inst_cbs:
            for cb in self._inst_cbs:
                cb(instr)
        
        r = self.ctx.processing(instr)

        # Set a unique comment on each symbolic expressions
        for e in instr.getSymbolicExpressions():
            e.setComment(self._fmt_comment())

        self._turn_off()
        return r

    def _fmt_comment(self) -> str:
        """Return a string identifying a SymbolicExpression in a unique manner"""
        return f"{self.inst_id}#{self.expr_id}#{self.current_address}"

    def add_instruction_callback(self, cb: Callable) -> None:
        """
        Add an Instruction callback that will be called for every
        instruction executed.
        """
        self._inst_cbs.append(cb)

    @staticmethod
    def _memacc_to_all_addr(ma: MemoryAccess) -> Set[int]:
        """Return a set of all addresses of a MemoryAccess"""
        addr = ma.getAddress()
        return set(range(addr, addr+ma.getSize()))

    @staticmethod
    def _split_unaligned_access(addr: Addr, size: ByteSize) -> List[MemoryAccess]:
        max_ma_size = 64  # Max aligned memory access size
        ma_size = 1
        splitted_ma = []
        current_addr = addr + size
        while ma_size < max_ma_size:
            if size & 0x1:
                current_addr -= ma_size
                splitted_ma.append(MemoryAccess(current_addr, ma_size))
            ma_size <<= 1
            size >>= 1
        for _ in range(size):
            current_addr -= ma_size
            splitted_ma.append(MemoryAccess(current_addr, ma_size))
        return splitted_ma

    @staticmethod
    def _coalesce_bytes_to_mas(ma_bytes: Iterable[Addr]) -> List[MemoryAccess]:
        """Convert a bunch of addresses into a list of MemoryAccess"""
        tmp_mem_accesses = []
        sorted_bytes = sorted(ma_bytes)
        addr, size = sorted_bytes[0], 1
        for b in sorted_bytes[1:]:
            if b == addr + size:
                size += 1
            else:
                tmp_mem_accesses.append((addr, size))
                addr = b
                size = 1
        tmp_mem_accesses.append((addr, size))
        coalesced_ma = []
        for addr, size in tmp_mem_accesses:
            coalesced_ma.extend(SimpleSymExec._split_unaligned_access(addr, size))
        return coalesced_ma



# built-in modules
from typing import Tuple, Dict

# qsynthesis modules
from qsynthesis.tritonast import TritonAst
from qsynthesis.algorithms.synthesizer_tdbu import TopDownBottomUpSynthesizer, YieldT, logger


class PlaceHolderSynthesizer(TopDownBottomUpSynthesizer):
    """
    Synthesizer inherited from TopDownBottomUpSynthesizer which thus
    form the search in that manner. The specificity of this synthesizer
    is to temporarily replacing synthesized expressions with a placeholder
    variable that will 'abstract the behavior of the synthesized AST'. The
    intended effect is to recursively synthesized previously synthesized
    sub-AST (with these placeholders). At the end of the search all placeholder
    variables are substituted by their associated synthesized expression.
    """

    def synthesize(self, ioast: TritonAst, check_sem: bool = False) -> Tuple[TritonAst, bool]:
        """
        Performs the placeholder based top-down and then bottom-up search for synthesizing
        the given TritonAst.

        :param ioast: TritonAst object to synthesize
        :param check_sem: boolean one whether to check the semantic equivalence of expression
                          before substituting them. That ensure soundness of the synthesis
        :return: tuple with new TritonAst and whether some replacement took place or not

        .. warning:: Activating the `check_sem` parameter implies a strong overhead
                     on the synthesis as SMT queries are being performed for any candidates
        """
        ioast = ioast.duplicate()  # Make a copy of the AST to to modify it
        self.expr_cache = {}
        self.eval_cache = {}
        self.call_to_eval = 0
        self.eval_count = 0

        expr_repl_visitor = self._visit_replacement_bfs(ioast)
        new_expr_to_send = None  # hold Ast replacement
        expr_modified = False    # True as soon as one sub-ast has been synthesized
        replacements = {}        # AST -> Placeholder AST
        replacements_hashs = {}  # AST hash -> Placeholder
        cur_placeholder = 0

        while 1:
            logger.debug(f"[plhdr] sending: {new_expr_to_send.pp_str if new_expr_to_send is not None else None}")
            cur_ast, info = expr_repl_visitor.send(new_expr_to_send)  # Iterate generator

            if isinstance(info, bool):
                final_ast = cur_ast
                break # Don't do anything of final expr return

            # Check if not already substituted, if so yield directly the placeholder
            if cur_ast.hash in replacements_hashs:
                logger.debug(f"[plhdr] Hash match: {cur_ast.pp_str} ==> {replacements_hashs[cur_ast.hash].pp_str}")
                new_expr_to_send = replacements_hashs[cur_ast.hash]
                continue

            # If don't have children either constant or variable just continue
            if not cur_ast.has_children(): # FIXME: If constant maybe symbolizing it
                continue

            # Try synthesizing expression
            logger.debug(f"[phldr] try synthesis lookup: {cur_ast.pp_str if cur_ast.node_count < 50 else 'too large'} [{cur_ast.node_count}] [{info.name}]")
            synt_res = self.try_synthesis_lookup(cur_ast, check_sem)

            if synt_res is not None:
                expr_modified = True
            else:
                new_expr_to_send = None
                if info == YieldT.TopBottom:
                    continue
                else:  # BottomUp
                    if self._ast_binop_with_cst(cur_ast):  # If binary operation with a constant make a placeholder
                        synt_res = cur_ast
                    else:
                        continue

            # Here necessarily synthesized of BottomUp

            # Create the a placeholder whether or not we synthesized the expression !
            placeholder = ioast.mk_variable(f"plhd_{cur_placeholder}", synt_res.size)
            cur_placeholder += 1
            replacements[placeholder] = synt_res
            replacements_hashs[cur_ast.hash] = placeholder
            new_expr_to_send = placeholder
            logger.debug(f"[plhdr] Create Placeholder plhd_{cur_placeholder-1} for: {synt_res.pp_str} ptr_id: {synt_res.ptr_id}")

        logger.debug(f"[plhdr] AST before replacement: {final_ast.pp_str}")
        for k, v in replacements.items():
            logger.debug(f"Final replace {k.pp_str}  ==> {v.pp_str}")
        self.replace_all(final_ast, replacements, recursive=True)
        final_ast.update_all()
        logger.info(f"Final AST: {final_ast.pp_str}")
        return final_ast, expr_modified

    @staticmethod
    def _ast_binop_with_cst(ast: TritonAst) -> bool:
        """
        Find if either the left or right branch of an AST is a constant.
        That enable replacing them with placeholder variable to help the
        synthesis by temporarily getting rid of constants.

        :param ast: TritonAst
        :return: True if left or right child of the AST is a constant
        """
        return sum(x.is_constant() for x in ast.get_children()) != 0

    def replace_all(self, ast: TritonAst, replacement: Dict[TritonAst, TritonAst], recursive: bool = False) -> None:
        """
        Performs the final replacement of Placeholder variable with their synthesized expression equivalence
        in `ast` provided in parameter.

        :param ast: TritonAst object in which to perform all substitutions
        :param replacement: dictionnary of placeholder variable as TritonAst to synthesized expression (TritonAst)
        :param recursive: whether to also perform the substitution in synthesized expressions given in the dictionnary
        :return: None as the `ast` object is modified in place

        .. warning:: All the expressions to replace should be present only once because TritonAst objects
                     are not deepcopied. Thus duplicating it at various location of the AST would be various dangerous
                     for attributes coherence when calling update
        """
        try:
            new_expr_to_send = None
            hash_mapping = {k.hash: k for k in replacement}

            logger.debug(f"Len replacement: {len(replacement)}  len hash_mapping:{len(hash_mapping)}")

            g = ast.visit_replacement(update=False)

            while True:
                cur_ast = g.send(new_expr_to_send)
                new_expr_to_send = None
                cur_ast_id = cur_ast.hash

                if cur_ast_id in hash_mapping:  # Prefilter with hash to check if current ast might be substituted
                    key = hash_mapping[cur_ast_id]
                    logger.debug(f"cur_ast: {cur_ast.pp_str}  key:{key.pp_str}  equal:{cur_ast.expr.equalTo(key.expr)}")

                    if not cur_ast.expr.equalTo(key.expr):  # Check that they are equal
                        logger.warning("two different expressions !")

                    new_expr_to_send = replacement[key]
                    logger.debug(f"Will replace {cur_ast.pp_str}  ==> {new_expr_to_send.pp_str}")

                    if recursive:
                        self.replace_all(new_expr_to_send, replacement, recursive=recursive)
        except StopIteration:
            pass



# built-in libs
from pathlib import Path
import hashlib
import json

# third-party libs
import plyvel

# qsynthesis deps
from qsynthesis.grammar import TritonGrammar, BvOp
from qsynthesis.tables.base import InputOutputOracle, HashType
from qsynthesis.types import Optional, List, Union, Tuple, Iterator, Hash, Input


META_KEY = b"metadatas"
VARS_KEY = b"variables"
INPUTS_KEY = b"inputs"
SIZE_KEY = b"size"


class InputOutputOracleLevelDB(InputOutputOracle):
    """
    Key-Value store oracle database based on Google Level-DB
    """
    def __init__(self, grammar: TritonGrammar, inputs: List[Input], hash_mode: HashType = HashType.RAW, f_name: str = ""):
        """
        Constructor making a InputOutputOracleLevelDB from a grammar a set of inputs and an hash type.

        :param grammar: triton grammar
        :param inputs: List of inputs
        :param hash_mode: type of hash to be used as keys in tables
        :param f_name: file name of the table (when being loaded)
        """
        super(InputOutputOracleLevelDB, self).__init__(grammar, inputs, hash_mode, f_name)
        self.db = None

    @staticmethod
    def create(filename: Union[str, Path], grammar: TritonGrammar, inputs: List[Input], hash_mode: HashType = HashType.RAW, constants: List[int] = []) -> 'InputOutputOracleLevelDB':
        """
        Create a new empty lookup table with the given initial parameters, grammars, inputs
        and hash_mode.

        :param filename: filename of the table to create
        :param grammar: TritonGrammar object representing variables and operators
        :param inputs: list of inputs on which to perform evaluation
        :param hash_mode: Hashing mode for keys
        :param constants: list of constants used
        :returns: InputOutputOracleLevelDB instance object
        """
        # TODO: If it exists deleting it ?
        db = plyvel.DB(str(filename), create_if_missing=True)

        metas = dict(hash_mode=hash_mode.name, operators=[x.name for x in grammar.ops], constants=constants)
        db.put(META_KEY, json.dumps(metas).encode())
        db.put(VARS_KEY, json.dumps(grammar.vars_dict).encode())
        db.put(INPUTS_KEY, json.dumps(inputs).encode())
        lkp = InputOutputOracleLevelDB(grammar=grammar, inputs=inputs, hash_mode=hash_mode, f_name=filename)
        lkp.db = db
        return lkp

    @staticmethod
    def load(file: Union[Path, str]) -> 'InputOutputOracleLevelDB':
        """
        Load the given lookup table and returns an instance object.

        :param file: Database file to load
        :returns: InputOutputOracleLevelDB object
        """
        db = plyvel.DB(str(file))
        metas = json.loads(db.get(META_KEY))
        ops = [BvOp[x] for x in metas['operators']]
        vrs = list(json.loads(db.get(VARS_KEY)).items())
        inps = json.loads(db.get(INPUTS_KEY))
        grammar = TritonGrammar(vars=vrs, ops=ops)
        lkp = InputOutputOracleLevelDB(grammar=grammar, inputs=inps, hash_mode=HashType[metas['hash_mode']], f_name=file)
        lkp.db = db
        return lkp

    def add_entry(self, hash: Hash, value: str) -> None:
        """
        Put the given hash and value in the leveldb trie.

        :param hash: already computed hash to add
        :param value: expression value to add in the table
        """
        self.db.put(hash, value.encode())
        self.db.put(SIZE_KEY, (str(int(self.db.get(SIZE_KEY))+1)).encode())

    def add_entries(self, entries: List[Tuple[Hash, str]], calc_hash: bool = False, chunk_size: int = 10000, update_count: bool = True) -> None:
        """
        Add the given list of entries in the database. The boolean ``calc_hash`` indicates
        whether hashes are already computed or not. If false the function should hash the
        hash first.

        :param entries: list of entries to add
        :param calc_hash: whether or not hash should be performed on entries keys
        :param chunk_size: size of a chunk for bulk insert in DB
        :param update_count: whether or not to update the count of entries in DB
        :returns: None
        """
        count = len(entries)

        def do_hash(x):
            return x if not calc_hash else (hashlib.md5(bytes(x)).digest() if self.hash_mode == HashType.MD5 else self.hash)

        for step in range(0, count, chunk_size):
            with self.db.write_batch(sync=True) as wb:
                for outs, s in entries[step:step+chunk_size]:
                    wb.put(do_hash(outs), s.encode())
        if update_count:
            cur_count = self.db.get(SIZE_KEY)
            new_count = count if cur_count is None else int(cur_count)+count
            self.db.put(SIZE_KEY, str(new_count).encode())

    def __iter__(self) -> Iterator[Tuple[Hash, str]]:
        """ Iterator of all the entries as an iterator of pair, hash, expression as string """
        for key, value in self.db:
            if key not in [META_KEY, VARS_KEY, INPUTS_KEY, SIZE_KEY]:
                yield key, value.decode()

    @property
    def is_writable(self) -> bool:
        """
        Whether the table enable being written (with new expressions)
        Level-db tables are considered to be always writable
        """
        return True

    @property
    def size(self) -> int:
        """Size of the table (number of entries)"""
        return int(self.db.get(SIZE_KEY))

    def _get_item(self, h: Hash) -> Optional[str]:
        """
        From a given hash return the associated expression string if
        found in the lookup table.

        :param h: hash of the item to get
        :returns: raw expression string if found
        """
        entry = self.db.get(h)
        return entry.decode() if entry else None



#!/usr/bin/env python3
# coding: utf-8
"""Installation script for qsynthesis module."""

from setuptools import setup, find_packages

generate_deps = ['sympy']
server_deps = ['fastapi', 'uvicorn']
assembly_deps = ["arybo", "llvmlite"]

setup(
    name="qsynthesis",
    version="0.1.1",
    description="Python API to synthesize Triton AST's",
    long_description_content_type='text/markdown',
    long_description="file: README.md",
    packages=find_packages(),
    url="https://github.com/quarkslab/qsynthesis",
    project_urls={
        "Documentation": "https://quarkslab.github.io/qsynthesis/",
        "Bug Tracker": "https://github.com/quarkslab/qsynthesis/issues",
        "Source": "https://github.com/quarkslab/qsynthesis"
    },
    setup_requires=[],
    install_requires=["triton-library",
                      "ordered_set",
                      "psutil",
                      "click",
                      "plyvel",
                      "requests",
                      "capstone",
                      "pydffi>=0.9.1"],
    tests_require=[],
    license="AGPL-3.0",
    author="Robin David",
    classifiers=[
        'Topic :: Security',
        'Environment :: Console',
        'Operating System :: OS Independent',
    ],
    extras_require={
        'all': assembly_deps+generate_deps+server_deps,
        'reassembly': assembly_deps,
        'generator': generate_deps,
        'server': server_deps
    },
    test_suite="",
    scripts=['bin/qsynthesis-table-manager', 'bin/qsynthesis-table-server']
)



from qsynthesis.algorithms import TopDownSynthesizer, PlaceHolderSynthesizer
from qsynthesis.grammar import TritonGrammar, BvOp
from qsynthesis.tables import InputOutputOracleREST, InputOutputOracleLevelDB, HashType
from qsynthesis.utils.symexec import SimpleSymExec
from qsynthesis.tritonast import TritonAst

import logging

__version__ = "0.1.1"

# Simple object used to retrieve the plugin in IDA
qsynthesis_plugin = None


def enable_logging(level: int = 0):
    logger = logging.getLogger("qsynthesis")
    if level:
        logger.setLevel(level)
    logger.disabled = False


def disable_logging():
    logger = logging.getLogger("qsynthesis")
    logger.disabled = True
